/*****************************************************************************
 * common.cpp - Common code used by all the sb16 miniports.
 *****************************************************************************
 * Copyright (c) 1997-2000 Microsoft Corporation.  All rights reserved.
 *
 * Implmentation of the common code object.  This class deals with interrupts
 * for the device, and is a collection of common code used by all the
 * miniports.
 */

#include "common.h"

#define STR_MODULENAME "sb16Adapter: "




/*****************************************************************************
 * CAdapterCommon

 *****************************************************************************
 * Adapter common object.
 */
class CAdapterCommon
:   public IAdapterCommon,
    public IAdapterPowerManagement,
    public CUnknown

{
private:
    PINTERRUPTSYNC          m_pInterruptSync;
    PPORTWAVECYCLIC         m_pPortWave;
    PSERVICEGROUP           m_pServiceGroupWave;
    PUCHAR                  m_pWaveBase;
    PUCHAR                  m_pAlsBase;
    ULONG                   Als4000PlaybackTransferCount;
    ULONG                   Als4000RecordTransferCount;
    BOOLEAN                 Als4000PlaybackStatus;
    BOOLEAN                 Als4000RecordStatus;
    ULONG                   ActualSampleRate;
    PWAVEMINIPORTSB16       m_WaveMiniportSB16;
    PDEVICE_OBJECT          m_pDeviceObject;
    DEVICE_POWER_STATE      m_PowerState;

    void AcknowledgeALLIRQ
    (   void
    );

public:
    DECLARE_STD_UNKNOWN();
    DEFINE_STD_CONSTRUCTOR(CAdapterCommon);
    ~CAdapterCommon();

    /*****************************************************************************
     * IAdapterCommon methods
     */
    STDMETHODIMP_(NTSTATUS) Init
    (
        IN      PRESOURCELIST   ResourceList,
        IN      PDEVICE_OBJECT  DeviceObject
    );
    STDMETHODIMP_(PINTERRUPTSYNC) GetInterruptSync
    (   void
    );
    STDMETHODIMP_(PUNKNOWN *) WavePortDriverDest
    (   void
    );
    STDMETHODIMP_(void) SetWaveServiceGroup
    (
        IN      PSERVICEGROUP   ServiceGroup
    );

    STDMETHODIMP_(void) SetWaveMiniport
    (
        IN      PWAVEMINIPORTSB16   Miniport
    );
    STDMETHODIMP_(BYTE) ReadController
    (   void
    );
    STDMETHODIMP_(BOOLEAN) WriteController
    (
        IN      BYTE    Value
    );
    STDMETHODIMP_(NTSTATUS) ResetController
    (   void
    );
    STDMETHODIMP_(void) MixerRegWrite
    (
        IN      BYTE    Index,
        IN      BYTE    Value
    );
    STDMETHODIMP_(BYTE) MixerRegRead
    (
        IN      BYTE    Index
    );
    STDMETHODIMP_(void) MixerReset
    (   void
    );
    STDMETHODIMP_(ULONG) GetActualSampleRateFromRegistry
    (   void
    );
    STDMETHODIMP_(BOOLEAN)GetLegacyGamePortEnableFlagFromRegistry
    (   void
    );
    STDMETHODIMP RestoreMixerSettingsFromRegistry
    (   void
    );
    STDMETHODIMP SaveMixerSettingsToRegistry
    (   void
    );
    STDMETHODIMP_(void) Als4000GcrRegWrite
    (
        IN      BYTE    Index,
        IN      ULONG   Value
    );
    STDMETHODIMP_(ULONG) Als4000GcrRegRead
    (
        IN      BYTE    Index
    );
    STDMETHODIMP_(void) Als4000RecordStart
    (
        IN      ULONG       Count,
        IN      BOOLEAN     FormatStereo
    );
    STDMETHODIMP_(void) Als4000RecordStop
    (
        IN      BOOLEAN     FormatStereo
    );
    STDMETHODIMP_(void) Als4000RecordPause
    (
        IN      BOOLEAN     FormatStereo
    );
    STDMETHODIMP_(BOOLEAN) Als4000RecordRestart
    (
        IN      BOOLEAN     FormatStereo
    );
    STDMETHODIMP_(void) Als4000PlaybackStart
    (
        IN      ULONG       Count,
        IN      BOOLEAN     FormatStereo
    );
    STDMETHODIMP_(void) Als4000PlaybackStop
    (
        IN      BOOLEAN     FormatStereo
    );
    STDMETHODIMP_(void) Als4000PlaybackPause
    (
        IN      BOOLEAN     FormatStereo
    );
    STDMETHODIMP_(BOOLEAN) Als4000PlaybackRestart
    (
        IN      BOOLEAN     FormatStereo
    );
    STDMETHODIMP_(void) Als4000SetPlaybackStartAddressAndCount
    (
        IN      ULONG       StartAddress,
        IN      ULONG       Count
    );
    STDMETHODIMP_(void) Als4000SetRecordStartAddressAndCount
    (
        IN      ULONG       StartAddress,
        IN      ULONG       Count
    );
    STDMETHODIMP_(ULONG) Als4000ReadPlaybackCounter
    (
        void
    );
    STDMETHODIMP_(ULONG) Als4000ReadRecordCounter
    (
        void
    );
    STDMETHODIMP_(ULONG) Als4000GetPlaybackTransferCount
    (
        void
    );
    STDMETHODIMP_(ULONG) Als4000GetRecordTransferCount
    (
        void
    );
    STDMETHODIMP_(void) Als4000SetupLegacyGamePort
    (
        IN      BOOLEAN     Enable
    );
    STDMETHODIMP_(ULONG)GetActualSampleRate
    (
        void
    );

    STDMETHODIMP_(void) Als4000Init
    (
        void
    );

    /*************************************************************************
     * IAdapterPowerManagement implementation
     *
     * This macro is from PORTCLS.H.  It lists all the interface's functions.
     */
    IMP_IAdapterPowerManagement;

    friend
    NTSTATUS
    InterruptServiceRoutine
    (
        IN      PINTERRUPTSYNC  InterruptSync,
        IN      PVOID           DynamicContext
    );

    friend
    BOOLEAN
    MixerRegReadSyncISR
    (
        PVOID Context
    );

    friend
    BOOLEAN
    MixerRegWriteSyncISR
    (
        PVOID Context
    );


};

static
MIXERSETTING MixerSettings[] =
{
    { L"LeftMasterVol",   DSP_MIX_MASTERVOLIDX_L,     0xF8, 0xF8 },
    { L"RightMasterVol",  DSP_MIX_MASTERVOLIDX_R,     0xF8, 0xF8 },
    { L"LeftWaveVol",     DSP_MIX_VOICEVOLIDX_L,      0xF8, 0xF8 },
    { L"RightWaveVol",    DSP_MIX_VOICEVOLIDX_R,      0xF8, 0xF8 },
    { L"LeftMidiVol",     DSP_MIX_FMVOLIDX_L,         0x00, 0x00 },
    { L"RightMidiVol",    DSP_MIX_FMVOLIDX_R,         0x00, 0x00 },
    { L"LeftCDVol",       DSP_MIX_CDVOLIDX_L,         0x00, 0x00 },
    { L"RightCDVol",      DSP_MIX_CDVOLIDX_R,         0x00, 0x00 },
    { L"LeftLineInVol",   DSP_MIX_LINEVOLIDX_L,       0x00, 0x00 },
    { L"RightLineInVol",  DSP_MIX_LINEVOLIDX_R,       0x00, 0x00 },
    { L"MicVol",          DSP_MIX_MICVOLIDX,          0x00, 0x00 },
    { L"PcSpkrVol",       DSP_MIX_SPKRVOLIDX,         0x00, 0x00 },
    { L"OutputMixer",     DSP_MIX_OUTMIXIDX,          0x00, 0x00 },
    { L"LeftInputMixer",  DSP_MIX_ADCMIXIDX_L,        0x00, 0x00 },
    { L"RightInputMixer", DSP_MIX_ADCMIXIDX_R,        0x00, 0x00 }
};


typedef struct {
    CAdapterCommon *that;
    ULONG   Length;
    BYTE    Data[4];
    ULONG   DWORDDATA;
} IOSYNCISR,*PIOSYNCISR;


#pragma code_seg("PAGE")

/*****************************************************************************
 * NewAdapterCommon()
 *****************************************************************************
 * Create a new adapter common object.
 */
NTSTATUS
NewAdapterCommon
(
    OUT     PUNKNOWN *  Unknown,
    IN      REFCLSID,
    IN      PUNKNOWN    UnknownOuter    OPTIONAL,
    IN      POOL_TYPE   PoolType
)
{
    PAGED_CODE();

    ASSERT(Unknown);

    STD_CREATE_BODY_
    (
        CAdapterCommon,
        Unknown,
        UnknownOuter,
        PoolType,
        PADAPTERCOMMON
    );
}


/*****************************************************************************
 * CAdapterCommon::Init()
 *****************************************************************************
 * Initialize an adapter common object.
 */
NTSTATUS
CAdapterCommon::
Init
(
    IN      PRESOURCELIST   ResourceList,
    IN      PDEVICE_OBJECT  DeviceObject
)
{
    PAGED_CODE();

    ASSERT(ResourceList);
    ASSERT(DeviceObject);

    //
    // Make sure we have the resources we expect
    //
    if ((ResourceList->NumberOfPorts() < 1) ||
        (ResourceList->NumberOfInterrupts() != 1))
    {
        _DbgPrintF (DEBUGLVL_TERSE, ("unknown configuration; check your code!"));
        // Bail out.
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    m_pDeviceObject = DeviceObject;
    m_WaveMiniportSB16 = NULL;

    //
    // Get the base address for the wave device.
    //
    ASSERT(ResourceList->FindTranslatedPort(0));
    m_pWaveBase = PUCHAR(ResourceList->FindTranslatedPort(0)->u.Port.Start.LowPart);
    m_pAlsBase  = m_pWaveBase;

    // SB Base
    m_pWaveBase += 0x10;

    // ALS4000̏
    Als4000Init();

    // WXgŎw肳ĂAPnPfoCXƂ
    // 200-201ɃQ[|[goB댯BłyB
    if(GetLegacyGamePortEnableFlagFromRegistry() == TRUE){
        Als4000SetupLegacyGamePort(TRUE);
    }

    // [U[ɂĎw肳ꂽۂ̃Tv[gWXg擾
    GetActualSampleRateFromRegistry();

    //
    // Set initial device power state
    //
    m_PowerState = PowerDeviceD0;

    // TEhuX^[Zbg
    NTSTATUS ntStatus = ResetController();

    if(NT_SUCCESS(ntStatus))
    {
        _DbgPrintF(DEBUGLVL_VERBOSE,("ResetController Succeeded"));
        AcknowledgeALLIRQ();

        //
        // Hook up the interrupt.
        //
        ntStatus = PcNewInterruptSync(                              // See portcls.h
                                        &m_pInterruptSync,          // Save object ptr
                                        NULL,                       // OuterUnknown(optional).
                                        ResourceList,               // He gets IRQ from ResourceList.
                                        0,                          // Resource Index
                                        InterruptSyncModeNormal     // Run ISRs once until we get SUCCESS
                                     );
        if (NT_SUCCESS(ntStatus) && m_pInterruptSync)
        {                                                                       //  run this ISR first
            ntStatus = m_pInterruptSync->RegisterServiceRoutine(InterruptServiceRoutine,PVOID(this),FALSE);
            if (NT_SUCCESS(ntStatus))
            {
                ntStatus = m_pInterruptSync->Connect();
            }

            // if we could not connect or register the ISR, release the object.
            if (!NT_SUCCESS (ntStatus))
            {
                m_pInterruptSync->Release();
                m_pInterruptSync = NULL;
            }
        }
    } else
    {
        _DbgPrintF(DEBUGLVL_TERSE,("ResetController Failure"));
    }

    return ntStatus;
}

/*****************************************************************************
 * CAdapterCommon::~CAdapterCommon()
 *****************************************************************************
 * Destructor.
 */
CAdapterCommon::
~CAdapterCommon
(   void
)
{
    PAGED_CODE();

    _DbgPrintF(DEBUGLVL_VERBOSE,("[CAdapterCommon::~CAdapterCommon]"));

    // TEhuX^[ZbgBĐ^^̋~B
    NTSTATUS ntStatus = ResetController();
    if(NT_SUCCESS(ntStatus))
    {
        _DbgPrintF(DEBUGLVL_VERBOSE,("ResetController Succeeded"));
    }

    // 200-201̃Q[|[g
    Als4000SetupLegacyGamePort(FALSE);

    if (m_pInterruptSync)
    {
        m_pInterruptSync->Disconnect();
        m_pInterruptSync->Release();
        m_pInterruptSync = NULL;
    }
    if (m_pPortWave)
    {
        m_pPortWave->Release();
    }
    if (m_pServiceGroupWave)
    {
        m_pServiceGroupWave->Release();
    }
}

/*****************************************************************************
 * CAdapterCommon::NonDelegatingQueryInterface()
 *****************************************************************************
 * Obtains an interface.
 */
STDMETHODIMP
CAdapterCommon::
NonDelegatingQueryInterface
(
    REFIID  Interface,
    PVOID * Object
)
{
    PAGED_CODE();

    ASSERT(Object);

    if (IsEqualGUIDAligned(Interface,IID_IUnknown))
    {
        *Object = PVOID(PUNKNOWN(PADAPTERCOMMON(this)));
    }
    else
    if (IsEqualGUIDAligned(Interface,IID_IAdapterCommon))
    {
        *Object = PVOID(PADAPTERCOMMON(this));
    }
    else
    if (IsEqualGUIDAligned(Interface,IID_IAdapterPowerManagment))
    {
        *Object = PVOID(PADAPTERPOWERMANAGMENT(this));
    }
    else
    {
        *Object = NULL;
    }

    if (*Object)
    {
        PUNKNOWN(*Object)->AddRef();
        return STATUS_SUCCESS;
    }

    return STATUS_INVALID_PARAMETER;
}

/*****************************************************************************
 * CAdapterCommon::GetInterruptSync()
 *****************************************************************************
 * Get a pointer to the interrupt synchronization object.
 */
STDMETHODIMP_(PINTERRUPTSYNC)
CAdapterCommon::
GetInterruptSync
(   void
)
{
    PAGED_CODE();

    return m_pInterruptSync;
}

/*****************************************************************************
 * CAdapterCommon::WavePortDriverDest()
 *****************************************************************************
 * Get a pointer to the wave port pointer.
 */
STDMETHODIMP_(PUNKNOWN *)
CAdapterCommon::
WavePortDriverDest
(   void
)
{
    PAGED_CODE();

    return (PUNKNOWN *) &m_pPortWave;
}



STDMETHODIMP_(void)
CAdapterCommon::
SetWaveMiniport
(
    IN      PWAVEMINIPORTSB16 Miniport
)
{
        m_WaveMiniportSB16 = Miniport;
}


/*****************************************************************************
 * CAdapterCommon::SetWaveServiceGroup()
 *****************************************************************************
 * Provides a pointer to the wave service group.
 */
STDMETHODIMP_(void)
CAdapterCommon::
SetWaveServiceGroup
(
    IN      PSERVICEGROUP   ServiceGroup
)
{
    if (m_pServiceGroupWave)
    {
        m_pServiceGroupWave->Release();
    }

    m_pServiceGroupWave = ServiceGroup;
    if( m_pServiceGroupWave )
    {
        m_pServiceGroupWave->AddRef();
    }
}


#pragma code_seg()

/*****************************************************************************
 * CAdapterCommon::ReadController()
 *****************************************************************************
 * Read a byte from the controller.
 */
STDMETHODIMP_(BYTE)
CAdapterCommon::
ReadController
(   void
)
{
    BYTE returnValue = BYTE(-1);

    ASSERT(m_pWaveBase);

    ULONGLONG startTime = PcGetTimeInterval(0);

    do
    {
        if
        (   READ_PORT_UCHAR
            (
                m_pWaveBase + DSP_REG_DATAAVAIL
            )
        &   0x80
        )
        {
            returnValue =
                READ_PORT_UCHAR
                (
                    m_pWaveBase + DSP_REG_READ
                );
        }
    } while ((PcGetTimeInterval(startTime) < GTI_MILLISECONDS(20)) &&
             (BYTE(-1) == returnValue));


    ASSERT((BYTE(-1) != returnValue) || !"ReadController timeout!");

    return returnValue;
}


/*****************************************************************************
 * CAdapterCommon::WriteController()
 *****************************************************************************
 * Write a byte to the controller.
 */
STDMETHODIMP_(BOOLEAN)
CAdapterCommon::
WriteController
(
    IN      BYTE    Value
)
{
    ASSERT(m_pWaveBase);

    BOOLEAN     returnValue = FALSE;
    ULONGLONG   startTime   = PcGetTimeInterval(0);

    do
    {

        BYTE status =
            READ_PORT_UCHAR
            (
                m_pWaveBase + DSP_REG_WRITE
            );

        if ((status & 0x80) == 0)
        {
            WRITE_PORT_UCHAR
            (
                m_pWaveBase + DSP_REG_WRITE,
                Value
            );

            returnValue = TRUE;
        }
    } while ((PcGetTimeInterval(startTime) < GTI_MILLISECONDS(20)) &&
              ! returnValue);


    ASSERT(returnValue || !"WriteController timeout");

    return returnValue;
}



/*
 *  MixerRegWrite()Ă΂AISRɊ荞܂Ȃ߂̃~LTWX^Cg[`B
 *  ISRŃ~LTWX^GȂȂ̂ŁAKv͖Ȃ̂܂܎cĂB
 */

BOOLEAN
MixerRegWriteSyncISR
(
    PVOID Context
)
{
    PIOSYNCISR iosync = (PIOSYNCISR)Context;
    CAdapterCommon *that = (CAdapterCommon *)iosync->that;

    WRITE_PORT_UCHAR (that->m_pWaveBase + DSP_REG_MIXREG,  iosync->Data[0]);
    WRITE_PORT_UCHAR (that->m_pWaveBase + DSP_REG_MIXDATA, iosync->Data[1]);
    return TRUE;
}


/*****************************************************************************
 * CAdapterCommon::MixerRegWrite()
 *****************************************************************************
 * Writes a mixer register.
 */
STDMETHODIMP_(void)
CAdapterCommon::
MixerRegWrite
(
    IN      BYTE    Index,
    IN      BYTE    Value
)
{
    ASSERT( m_pWaveBase );

    // only hit the hardware if we're in an acceptable power state
    if( m_PowerState <= PowerDeviceD1 )
    {
        IOSYNCISR iosync;
        PINTERRUPTSYNC Sync;
        BYTE Result;

        iosync.that = this;
        iosync.Length = 2;
        iosync.Data[0] = Index;
        iosync.Data[1] = Value;

        Sync = GetInterruptSync();
        if(Sync){
            KeSynchronizeExecution(Sync->GetKInterrupt(), MixerRegWriteSyncISR, &iosync);
            Result = iosync.Data[0];
        }else{
            WRITE_PORT_UCHAR(m_pWaveBase + DSP_REG_MIXREG,  Index);
            WRITE_PORT_UCHAR(m_pWaveBase + DSP_REG_MIXDATA, Value);
        }
    }

    // save current setting
    for(ULONG i = 0; i < SIZEOF_ARRAY(MixerSettings); i++)
    {
        if( Index == MixerSettings[i].RegisterIndex){
            MixerSettings[i].CurrentSetting = Value;
            break;
        }
    }
}


/*
 *  MixerRegRead()Ă΂AISRɊ荞܂Ȃ߂̃~LTWX^[h[`B
 *  ISRŃ~LTWX^GȂȂ̂ŁAKv͖Ȃ̂܂܎cĂB
 */

BOOLEAN
MixerRegReadSyncISR
(
    PVOID Context
)
{
    PIOSYNCISR iosync = (PIOSYNCISR)Context;
    CAdapterCommon *that = (CAdapterCommon *)iosync->that;

    WRITE_PORT_UCHAR (that->m_pWaveBase + DSP_REG_MIXREG, iosync->Data[0]);
    iosync->Data[0] = READ_PORT_UCHAR (that->m_pWaveBase + DSP_REG_MIXDATA);
    return TRUE;
}


/*****************************************************************************
 * CAdapterCommon::MixerRegRead()
 *****************************************************************************
 * Reads a mixer register.
 */
STDMETHODIMP_(BYTE)
CAdapterCommon::
MixerRegRead
(
    IN      BYTE    Index
)
{
    IOSYNCISR iosync;
    PINTERRUPTSYNC Sync;
    BYTE Result;

    iosync.that = this;
    iosync.Length = 1;
    iosync.Data[0] = Index;

    Sync = GetInterruptSync();

    if(Sync){
        KeSynchronizeExecution(Sync->GetKInterrupt(), MixerRegReadSyncISR, &iosync);
        Result = iosync.Data[0];
    }else{
        WRITE_PORT_UCHAR (m_pWaveBase + DSP_REG_MIXREG, Index);
        Result = READ_PORT_UCHAR (m_pWaveBase + DSP_REG_MIXDATA);
    }
    return Result;
}




/*****************************************************************************
 * CAdapterCommon::MixerReset()
 *****************************************************************************
 * Resets the mixer
 */
STDMETHODIMP_(void)
CAdapterCommon::
MixerReset
(   void
)
{
    ASSERT(m_pWaveBase);

    WRITE_PORT_UCHAR(m_pWaveBase + DSP_REG_MIXREG, DSP_MIX_DATARESETIDX);
    WRITE_PORT_UCHAR(m_pWaveBase + DSP_REG_MIXDATA, 0);

    MixerRegWrite(DSP_MIX_INGAINIDX_L, 0x40);       // input gain *2
    MixerRegWrite(DSP_MIX_INGAINIDX_R, 0x40);       // input gain *2
    MixerRegWrite(DSP_MIX_OUTGAINIDX_L, 0x40);      // output gain *2
    MixerRegWrite(DSP_MIX_OUTGAINIDX_R, 0x40);      // output gain *2

    MixerRegWrite(DSP_MIX_MONOCTRLIDX, 0x80);       // mono control
    MixerRegWrite(DSP_MIX_OUTMIX2IDX, 0x06);        // digital audio D/A enable
    MixerRegWrite(DSP_MIX_MICGAINIDX, 0x03);        // microphone 20dB gain
    MixerRegWrite(DSP_MIX_LPFCTRLIDX, 0x20);        // D/A 20KHz LPF
    MixerRegWrite(DSP_MIX_3DCTRLIDX,  0xA4);        // 3D effect disable

    RestoreMixerSettingsFromRegistry();
}

/*****************************************************************************
 * CAdapterCommon::AcknowledgeALLIRQ()
 *****************************************************************************
 * Acknowledge interrupt request.
 */
void
CAdapterCommon::
AcknowledgeALLIRQ
(   void
)
{
    ASSERT(m_pWaveBase);
    ASSERT(m_pAlsBase);

    READ_PORT_UCHAR(m_pWaveBase + DSP_REG_ACK16BIT);        // BX command      (playback 16bit)
    READ_PORT_UCHAR(m_pWaveBase + DSP_REG_ACK8BIT);         // non-BX command  (playback 8bit)
    READ_PORT_UCHAR(m_pWaveBase + DSP_REG_RESET);           // CR1E            (record)
    READ_PORT_UCHAR(m_pAlsBase  + ALS_REG_MIDI_DATA);       // MPU401
}

/*****************************************************************************
 * CAdapterCommon::ResetController()
 *****************************************************************************
 * Resets the controller.
 */

STDMETHODIMP_(NTSTATUS)
CAdapterCommon::
ResetController(void)
{
    NTSTATUS ntStatus = STATUS_UNSUCCESSFUL;

    if(m_pWaveBase && m_pAlsBase){
        ULONG temp = Als4000GcrRegRead(0x8C);
        Als4000GcrRegWrite(0x8C, temp & 0xFFFF7FFF);        // disable INTA#

        // write a 1 to the reset bit
        WRITE_PORT_UCHAR(m_pWaveBase + DSP_REG_RESET,1);

        // wait for  at least 3 microseconds
        KeStallExecutionProcessor( 5L );    // okay, 5us

        // write a 0 to the reset bit
        WRITE_PORT_UCHAR(m_pWaveBase + DSP_REG_RESET,0);

        // hang out for 100us
        KeStallExecutionProcessor( 100L );

        // read the controller
        BYTE ReadVal = ReadController();

        // check return value
        if( ReadVal == BYTE(0xAA) )
        {
            ntStatus = STATUS_SUCCESS;
        }
        Als4000GcrRegWrite(0x8C, temp);                     // restore
    }

    Als4000PlaybackStatus = FALSE;
    Als4000RecordStatus = FALSE;

    return ntStatus;
}

/*****************************************************************************
 * CAdapterCommon::Als4000GcrRegWrite()
 *****************************************************************************
 * Writes a ALS4000 GCR register.
 */

STDMETHODIMP_(void)
CAdapterCommon::
Als4000GcrRegWrite
(
    IN      BYTE    Index,
    IN      ULONG   Value
)
{
    ASSERT( m_pAlsBase );

    // only hit the hardware if we're in an acceptable power state
    if( m_PowerState <= PowerDeviceD1 )
    {
        WRITE_PORT_UCHAR
        (
            m_pAlsBase + 0x0C,
            Index
        );

        WRITE_PORT_ULONG
        (
            (PULONG)(m_pAlsBase + 0x08),
            Value
        );
    }
}


/*****************************************************************************
 * CAdapterCommon::Als4000GcrRegRead()
 *****************************************************************************
 * Reads a ALS4000 GCR register.
 */

STDMETHODIMP_(ULONG)
CAdapterCommon::
Als4000GcrRegRead
(
    IN      BYTE    Index
)
{
    ASSERT( m_pAlsBase );

    WRITE_PORT_UCHAR(m_pAlsBase + 0x0C, Index);
    return READ_PORT_ULONG( (PULONG)(m_pAlsBase + 0x08) );
}



/*****************************************************************************
 * CAdapterCommon::Als4000RecordStart()
 *****************************************************************************
 * ^Jn܂B
 */

STDMETHODIMP_(void)
CAdapterCommon::
Als4000RecordStart
(
    IN      ULONG       Count,
    IN      BOOLEAN     FormatStereo
)
{
    BYTE Mode;

    _DbgPrintF(DEBUGLVL_VERBOSE,("[Als4000RecordStart]"));
    _DbgPrintF(DEBUGLVL_VERBOSE,("BlockLength %d", Count));

    Count -= 1;
    MixerRegWrite(0xC0 | 0x1C, (BYTE)Count);            // low byte
    MixerRegWrite(0xC0 | 0x1D, (BYTE)(Count >> 8));     // high byte

    if(FormatStereo){
        // signed 16bit stereo format
        Mode = 0xB0;
    }else{
        // signed 16bit mono format
        Mode = 0x90;
    }
    // recording start
    MixerRegWrite(0xC0 | 0x1E, Mode);

    Als4000RecordStatus = TRUE;
}


/*****************************************************************************
 * CAdapterCommon::Als4000RecordStop()
 *****************************************************************************
 * ^~܂B
 */

STDMETHODIMP_(void)
CAdapterCommon::
Als4000RecordStop
(
    IN      BOOLEAN     FormatStereo
)
{
    BYTE Mode;

    _DbgPrintF(DEBUGLVL_VERBOSE,("[Als4000RecordStop]"));

    if(FormatStereo){
        // signed 16bit stereo format
        Mode = 0x30;
    }else{
        // signed 16bit mono format
        Mode = 0x10;
    }

    // recording stop
    MixerRegWrite(0xC0 | 0x1E, Mode);

    Als4000RecordStatus = FALSE;
}


/*****************************************************************************
 * CAdapterCommon::Als4000RecordPause()
 *****************************************************************************
 * ^̈ꎞ~B
 */

STDMETHODIMP_(void)
CAdapterCommon::
Als4000RecordPause
(
    IN      BOOLEAN     FormatStereo
)
{
    BYTE Mode;

    _DbgPrintF(DEBUGLVL_VERBOSE,("[Als4000RecordPause]"));

    if(FormatStereo){
        // signed 16bit stereo format
        Mode = 0xF0;
    }else{
        // signed 16bit mono format
        Mode = 0xD0;
    }
    // recording pause
    MixerRegWrite(0xC0 | 0x1E, Mode);
}

/*****************************************************************************
 * CAdapterCommon::Als4000RecordRestart()
 *****************************************************************************
 * ^̈ꎞ~̍ĊJB
 */

STDMETHODIMP_(BOOLEAN)
CAdapterCommon::
Als4000RecordRestart
(
    IN      BOOLEAN     FormatStereo
)
{
    BYTE Mode;
    BOOLEAN Status;

    _DbgPrintF(DEBUGLVL_VERBOSE,("[Als4000RecordRestart]"));

    if(Als4000RecordStatus == FALSE)
    {
        Status = FALSE;
    }
    else
    {
        if(FormatStereo){
            // signed 16bit stereo format
            Mode = 0xB0;
        }else{
            // signed 16bit mono format
            Mode = 0x90;
        }
        // recording restart
        MixerRegWrite(0xC0 | 0x1E, Mode);
        Status = TRUE;
    }
    return Status;
}


/*****************************************************************************
 * CAdapterCommon::Als4000PlaybackStart()
 *****************************************************************************
 * ĐJn܂B
 */

STDMETHODIMP_(void)
CAdapterCommon::
Als4000PlaybackStart
(
    IN      ULONG       Count,
    IN      BOOLEAN     FormatStereo
)
{
    BYTE Mode;

    _DbgPrintF(DEBUGLVL_VERBOSE,("[Als4000PlaybackStart]"));
    _DbgPrintF(DEBUGLVL_VERBOSE,("BlockLength %d", Count));

    WriteController(DSP_CMD_SPKRON);
    // 16bit format
    WriteController(DSP_CMD_STARTDAC16);
    Mode = 0x10;
    if (FormatStereo){
        Mode |= 0x20;
    }
    WriteController(Mode) ;
    Count -= 1;
    WriteController((BYTE) Count) ;
    WriteController((BYTE) (Count >> 8));

    Als4000PlaybackStatus = TRUE;
}


/*****************************************************************************
 * CAdapterCommon::Als4000PlaybackStop()
 *****************************************************************************
 * Đ~܂B
 */

STDMETHODIMP_(void)
CAdapterCommon::
Als4000PlaybackStop
(
    IN      BOOLEAN     FormatStereo
)
{
    _DbgPrintF(DEBUGLVL_VERBOSE,("[Als4000PlaybackStop]"));

    WriteController(DSP_CMD_PAUSEDMA16);
    WriteController(DSP_CMD_HALTAUTODMA16);
    WriteController(DSP_CMD_PAUSEDMA16);

    Als4000PlaybackStatus = FALSE;
}


/*****************************************************************************
 * CAdapterCommon::Als4000PlaybackPause()
 *****************************************************************************
 * Đ̈ꎞ~B
 */

STDMETHODIMP_(void)
CAdapterCommon::
Als4000PlaybackPause
(
    IN      BOOLEAN     FormatStereo
)
{
    _DbgPrintF(DEBUGLVL_VERBOSE,("[Als4000PlaybackPause]"));

    WriteController(DSP_CMD_PAUSEDMA16);
}


/*****************************************************************************
 * CAdapterCommon::Als4000PlaybackRestart()
 *****************************************************************************
 * Đ̈ꎞ~̍ĊJB
 */

STDMETHODIMP_(BOOLEAN)
CAdapterCommon::
Als4000PlaybackRestart
(
    IN      BOOLEAN     FormatStereo
)
{
    BOOLEAN Status;

    _DbgPrintF(DEBUGLVL_VERBOSE,("[Als4000PlaybackRestart]"));

    if(Als4000PlaybackStatus == FALSE)
    {
        Status = FALSE;
    }
    else
    {
        WriteController(DSP_CMD_CONTDMA16);
        Status = TRUE;
    }
    return Status;
}


/*****************************************************************************
 * CAdapterCommon::Als4000SetPlaybackStartAddressAndCount()
 *****************************************************************************
 * Đobt@AhXݒ
 */

STDMETHODIMP_(void)
CAdapterCommon::
Als4000SetPlaybackStartAddressAndCount
(
    IN      ULONG   StartAddress,
    IN      ULONG   Count
)
{
    _DbgPrintF(DEBUGLVL_VERBOSE,("[Als4000SetPlaybackStartAddressAndCount] base %x, addr %x, count %x",
        (ULONG)m_pAlsBase, (ULONG)StartAddress, (ULONG)Count));

    Als4000GcrRegWrite(0x91, StartAddress);
    Als4000GcrRegWrite(0x92, (Count - 1) | (1 << 20) | (1 << 19));
    Als4000PlaybackTransferCount = Count;

    Als4000GcrRegWrite(0x99, Als4000GcrRegRead(0x99) & 0x0FFFFFFF);
}

/*****************************************************************************
 * CAdapterCommon::Als4000SetRecordStartAddressAndCount()
 *****************************************************************************
 * ^obt@AhXݒ
 */

STDMETHODIMP_(void)
CAdapterCommon::
Als4000SetRecordStartAddressAndCount
(
    IN      ULONG   StartAddress,
    IN      ULONG   Count
)
{
    _DbgPrintF(DEBUGLVL_VERBOSE,("[Als4000SetRecordStartAddressAndCount] base %x, addr %x, count %x",
        (ULONG)m_pAlsBase, (ULONG)StartAddress, (ULONG)Count));

    Als4000GcrRegWrite(0xA2, StartAddress);
    Als4000GcrRegWrite(0xA3, Count - 1);
    Als4000RecordTransferCount = Count;

    Als4000GcrRegWrite(0x99, Als4000GcrRegRead(0x99) & 0x0FFFFFFF);
}


/*****************************************************************************
 * CAdapterCommon::Als4000ReadPlaybackCounter()
 *****************************************************************************
 * ݂̍ĐJE^iDMA]cʁj擾
 */

STDMETHODIMP_(ULONG)
CAdapterCommon::
Als4000ReadPlaybackCounter
(
    void
)
{
    ULONG Count;

    if(Als4000PlaybackStatus == TRUE){
        Count = Als4000GcrRegRead(0xA1) & 0xFFFF;
    }else{
        if(Als4000PlaybackTransferCount != 0){
            Count = Als4000PlaybackTransferCount - 1;
        }else{
            Count = Als4000PlaybackTransferCount;
        }
    }
    return Count;
}


/*****************************************************************************
 * CAdapterCommon::Als4000ReadRecordCounter()
 *****************************************************************************
 * ݂̘^JE^iDMA]cʁj擾
 */

STDMETHODIMP_(ULONG)
CAdapterCommon::
Als4000ReadRecordCounter
(
    void
)
{
    ULONG Count;

    if(Als4000RecordStatus == TRUE){
        Count = Als4000GcrRegRead(0xA5) & 0xFFFF;
    }else{
        if(Als4000RecordTransferCount != 0){
            Count = Als4000RecordTransferCount - 1;
        }else{
            Count = Als4000RecordTransferCount;
        }
    }
    return Count;
}


/*****************************************************************************
 * CAdapterCommon::Als4000GetPlaybackTransferCount()
 *****************************************************************************
 * Đobt@̒擾
 */

STDMETHODIMP_(ULONG)
CAdapterCommon::
Als4000GetPlaybackTransferCount
(
    void
)
{
    return Als4000PlaybackTransferCount;
}


/*****************************************************************************
 * CAdapterCommon::Als4000GetRecordTransferCount()
 *****************************************************************************
 * ^obt@̒擾
 */

STDMETHODIMP_(ULONG)
CAdapterCommon::
Als4000GetRecordTransferCount
(
    void
)
{
    return Als4000RecordTransferCount;
}


/*****************************************************************************
 * CAdapterCommon::GetActualSampleRate()
 *****************************************************************************
 * WXgŎw肳ꂽAۂ̃Tv[g擾
 */

STDMETHODIMP_(ULONG)
CAdapterCommon::
GetActualSampleRate
(
    void
)
{
    return ActualSampleRate;
}


/*****************************************************************************
 * CAdapterCommon::Als4000Init()
 *****************************************************************************
 * ALS4000n[hEFȀ
 */

STDMETHODIMP_(void)
CAdapterCommon::
Als4000Init
(
    void
)
{
    BYTE  temp;
    ULONG temp2;

    _DbgPrintF (DEBUGLVL_TERSE, ("Initialize ALS4000 Hardware"));

    temp = MixerRegRead(0xC0);
    MixerRegWrite(0xC0, temp | 0x80);       // disable mx80/81 write protection
    MixerRegWrite(0x81, 0x01);              // select DMA channel 0
    MixerRegWrite(0xC0, temp & 0x7F);       // enable mx80/81 write protection

    temp2 = Als4000GcrRegRead(0x8C);
    temp2 &= ~(1 << 27);                    // INTA# level trigger mode
    temp2 |= (1 << 15);                     // enable IRQ
    Als4000GcrRegWrite(0x8C, temp2);

    temp2 = Als4000GcrRegRead(0x99);
    temp2 &= 0xFFFFFFF8;
    temp2 |= (1 << 2);                      // burst mode for playback enable
    temp2 |= (1 << 16);                     // legacy dma emulation disable
    temp2 &= ~(1 << 20);                    // REQ# generated when PCM buffer is half empty
    Als4000GcrRegWrite(0x99, temp2);

    Als4000GcrRegWrite(0xA8, 0);            // disable legacy gameport & adlib
    Als4000GcrRegWrite(0xA9, 0);            // disable legacy mpu401 & sb

    Als4000GcrRegWrite(0x91, 0);            // clear dma 0 starting address
    Als4000GcrRegWrite(0x92, 0);            // clear dma 0 base byte count

    MixerReset();
}


/*****************************************************************************
 * CAdapterCommon::Als4000SetupLegacyGamePort()
 *****************************************************************************
 * 201h̃KV[Q[|[g̐ݒ
 */

STDMETHODIMP_(void)
CAdapterCommon::
Als4000SetupLegacyGamePort
(
    IN  BOOLEAN     Enable
)
{
    ULONG   Data = 0x200;                   // legacy gameport base I/O address
                                            // using 200h and 201h
    Data = Data << 16;
    if(Enable){
        Data |= 0x01 << 16;                 // set legacy gameport enable bit
        Als4000GcrRegWrite(0xA8, Data);     // enable legacy gameport
    }else{
        Als4000GcrRegWrite(0xA8, Data);     // disable legacy gameport
    }
}


/*****************************************************************************
 * CAdapterCommon::GetLegacyGamePortEnableFlagFromRegistry()
 *****************************************************************************
 * 201hɃQ[|[go邩ǂ̐ݒAWXgǂ
 */

STDMETHODIMP_(BOOLEAN)
CAdapterCommon::
GetLegacyGamePortEnableFlagFromRegistry
(   void
)
{
    PREGISTRYKEY    DriverKey;
    PREGISTRYKEY    ConfigKey;
    BYTE            Flag = 0;

    _DbgPrintF(DEBUGLVL_VERBOSE,("[GetLegacyGamePortEnableFlagFromRegistry]"));

    // open the driver registry key
    NTSTATUS ntStatus = PcNewRegistryKey( &DriverKey,               // IRegistryKey
                                          NULL,                     // OuterUnknown
                                          DriverRegistryKey,        // Registry key type
                                          KEY_ALL_ACCESS,           // Access flags
                                          m_pDeviceObject,          // Device object
                                          NULL,                     // Subdevice
                                          NULL,                     // ObjectAttributes
                                          0,                        // Create options
                                          NULL );                   // Disposition
    if(NT_SUCCESS(ntStatus))
    {
        UNICODE_STRING  KeyName;
        UNICODE_STRING  GamePortKeyName;
        ULONG           Disposition;

        // make a unicode string for the subkey name
        RtlInitUnicodeString( &KeyName, L"Config" );
        RtlInitUnicodeString( &GamePortKeyName, L"LegacyGamePort201hEnable" );

        // open the settings subkey
        ntStatus = DriverKey->NewSubKey( &ConfigKey,                // Subkey
                                         NULL,                      // OuterUnknown
                                         KEY_ALL_ACCESS,            // Access flags
                                         &KeyName,                  // Subkey name
                                         REG_OPTION_NON_VOLATILE,   // Create options
                                         &Disposition );
        if(NT_SUCCESS(ntStatus))
        {
            ULONG   ResultLength;

            if(Disposition == REG_CREATED_NEW_KEY)
            {
                _DbgPrintF(DEBUGLVL_VERBOSE,("REG_CREATED_NEW_KEY"));
                // set the key
                ConfigKey->SetValueKey( &GamePortKeyName,           // Key name
                                        REG_BINARY,                 // Key type
                                        PVOID(&Flag),
                                        sizeof(BYTE) );
            } else
            {
                // allocate data to hold key info
                PVOID KeyInfo = ExAllocatePool(PagedPool, sizeof(KEY_VALUE_PARTIAL_INFORMATION) + sizeof(BYTE));
                if(NULL != KeyInfo)
                {
                    // query the value key
                    ntStatus = ConfigKey->QueryValueKey( &GamePortKeyName,
                                                           KeyValuePartialInformation,
                                                           KeyInfo,
                                                           sizeof(KEY_VALUE_PARTIAL_INFORMATION) + sizeof(BYTE),
                                                           &ResultLength );
                    if(NT_SUCCESS(ntStatus))
                    {
                        PKEY_VALUE_PARTIAL_INFORMATION PartialInfo = PKEY_VALUE_PARTIAL_INFORMATION(KeyInfo);
                        if(PartialInfo->DataLength == sizeof(BYTE)){
                            Flag = *(PBYTE(PartialInfo->Data));
                            _DbgPrintF(DEBUGLVL_VERBOSE,("LegacyGamePortFlag from Registry"));
                        }
                    } else
                    {
                        _DbgPrintF(DEBUGLVL_VERBOSE,("key access failed"));
                        _DbgPrintF(DEBUGLVL_VERBOSE,("REG_CREATED_NEW_KEY"));
                        // if key access failed, set the default
                        ConfigKey->SetValueKey( &GamePortKeyName,       // Key name
                                                REG_BINARY,             // Key type
                                                PVOID(&Flag),
                                                sizeof(BYTE) );
                    }
                    // free the key info
                    ExFreePool(KeyInfo);
                }
            }
            // release the settings key
            ConfigKey->Release();
        }
        // release the driver key
        DriverKey->Release();
    }

    _DbgPrintF(DEBUGLVL_VERBOSE,("LegacyGamePortFlag %s", Flag ? "Enable" : "Disable"));

    return Flag ? TRUE : FALSE;
}


/*****************************************************************************
 * CAdapterCommon::GetActualSampleRateFromRegistry()
 *****************************************************************************
 * ۂ̃Tv[g̒lWXgǂށBftHg48211HzB
 */

STDMETHODIMP_(ULONG)
CAdapterCommon::
GetActualSampleRateFromRegistry
(   void
)
{
    PREGISTRYKEY    DriverKey;
    PREGISTRYKEY    ConfigKey;
    ULONG SampleRate = ALS4000_DEFAULT_RATE;

    _DbgPrintF(DEBUGLVL_VERBOSE,("[GetActualRateFromRegistry]"));

    // open the driver registry key
    NTSTATUS ntStatus = PcNewRegistryKey( &DriverKey,               // IRegistryKey
                                          NULL,                     // OuterUnknown
                                          DriverRegistryKey,        // Registry key type
                                          KEY_ALL_ACCESS,           // Access flags
                                          m_pDeviceObject,          // Device object
                                          NULL,                     // Subdevice
                                          NULL,                     // ObjectAttributes
                                          0,                        // Create options
                                          NULL );                   // Disposition
    if(NT_SUCCESS(ntStatus))
    {
        UNICODE_STRING  KeyName;
        UNICODE_STRING  ActualSampleRateKeyName;
        ULONG           Disposition;

        // make a unicode string for the subkey name
        RtlInitUnicodeString( &KeyName, L"Config" );
        RtlInitUnicodeString( &ActualSampleRateKeyName, L"ActualSampleRate" );

        // open the settings subkey
        ntStatus = DriverKey->NewSubKey( &ConfigKey,                // Subkey
                                         NULL,                      // OuterUnknown
                                         KEY_ALL_ACCESS,            // Access flags
                                         &KeyName,                  // Subkey name
                                         REG_OPTION_NON_VOLATILE,   // Create options
                                         &Disposition );
        if(NT_SUCCESS(ntStatus))
        {
            ULONG   ResultLength;

            if(Disposition == REG_CREATED_NEW_KEY)
            {
                _DbgPrintF(DEBUGLVL_VERBOSE,("REG_CREATED_NEW_KEY"));
                // set the key
                DWORD KeyValue = ALS4000_DEFAULT_RATE;
                ConfigKey->SetValueKey( &ActualSampleRateKeyName,   // Key name
                                        REG_DWORD,                  // Key type
                                        PVOID(&KeyValue),
                                        sizeof(DWORD) );
            } else
            {
                // allocate data to hold key info
                PVOID KeyInfo = ExAllocatePool(PagedPool, sizeof(KEY_VALUE_PARTIAL_INFORMATION) + sizeof(DWORD));
                if(NULL != KeyInfo)
                {
                    // init key name
                    RtlInitUnicodeString( &KeyName, L"ActualSampleRate" );

                    // query the value key
                    ntStatus = ConfigKey->QueryValueKey( &KeyName,
                                                           KeyValuePartialInformation,
                                                           KeyInfo,
                                                           sizeof(KEY_VALUE_PARTIAL_INFORMATION) + sizeof(DWORD),
                                                           &ResultLength );
                    if(NT_SUCCESS(ntStatus))
                    {
                        PKEY_VALUE_PARTIAL_INFORMATION PartialInfo = PKEY_VALUE_PARTIAL_INFORMATION(KeyInfo);
                        if(PartialInfo->DataLength == sizeof(DWORD)){
                            SampleRate = *(PDWORD(PartialInfo->Data));
                            _DbgPrintF(DEBUGLVL_VERBOSE,("ActualRate rate from Registry"));
                        }
                    } else
                    {
                        _DbgPrintF(DEBUGLVL_VERBOSE,("key access failed"));
                        _DbgPrintF(DEBUGLVL_VERBOSE,("REG_CREATED_NEW_KEY"));
                        // if key access failed, set the default
                        DWORD KeyValue = ALS4000_DEFAULT_RATE;
                        ConfigKey->SetValueKey( &ActualSampleRateKeyName,   // Key name
                                                REG_DWORD,                  // Key type
                                                PVOID(&KeyValue),
                                                sizeof(DWORD) );
                    }
                    // free the key info
                    ExFreePool(KeyInfo);
                }
            }
            // release the settings key
            ConfigKey->Release();
        }
        // release the driver key
        DriverKey->Release();
    }

    _DbgPrintF(DEBUGLVL_VERBOSE,("ActualSampleRate %d Hz", SampleRate));

    ActualSampleRate = SampleRate;
    return SampleRate;
}




/*****************************************************************************
 * CAdapterCommon::RestoreMixerSettingsFromRegistry()
 *****************************************************************************
 * Restores the mixer settings based on settings stored in the registry.
 */
STDMETHODIMP
CAdapterCommon::
RestoreMixerSettingsFromRegistry
(   void
)
{
    PREGISTRYKEY    DriverKey;
    PREGISTRYKEY    SettingsKey;

    _DbgPrintF(DEBUGLVL_VERBOSE,("[RestoreMixerSettingsFromRegistry]"));

    // open the driver registry key
    NTSTATUS ntStatus = PcNewRegistryKey( &DriverKey,               // IRegistryKey
                                          NULL,                     // OuterUnknown
                                          DriverRegistryKey,        // Registry key type
                                          KEY_ALL_ACCESS,           // Access flags
                                          m_pDeviceObject,          // Device object
                                          NULL,                     // Subdevice
                                          NULL,                     // ObjectAttributes
                                          0,                        // Create options
                                          NULL );                   // Disposition
    if(NT_SUCCESS(ntStatus))
    {
        UNICODE_STRING  KeyName;
        ULONG           Disposition;

        // make a unicode strong for the subkey name
        RtlInitUnicodeString( &KeyName, L"Settings" );

        // open the settings subkey
        ntStatus = DriverKey->NewSubKey( &SettingsKey,              // Subkey
                                         NULL,                      // OuterUnknown
                                         KEY_ALL_ACCESS,            // Access flags
                                         &KeyName,                  // Subkey name
                                         REG_OPTION_NON_VOLATILE,   // Create options
                                         &Disposition );
        if(NT_SUCCESS(ntStatus))
        {
            ULONG   ResultLength;

            if(Disposition == REG_CREATED_NEW_KEY)
            {
                _DbgPrintF(DEBUGLVL_VERBOSE,("REG_CREATED_NEW_KEY"));
                // copy default settings
                for(ULONG i = 0; i < SIZEOF_ARRAY(MixerSettings); i++)
                {
                    MixerRegWrite( MixerSettings[i].RegisterIndex,
                                   MixerSettings[i].DefaultSetting );
                }
                SaveMixerSettingsToRegistry();
            } else
            {
                // allocate data to hold key info
                PVOID KeyInfo = ExAllocatePool(PagedPool, sizeof(KEY_VALUE_PARTIAL_INFORMATION) + sizeof(DWORD));
                if(NULL != KeyInfo)
                {
                    // loop through all mixer settings
                    for(UINT i = 0; i < SIZEOF_ARRAY(MixerSettings); i++)
                    {
                        // init key name
                        RtlInitUnicodeString( &KeyName, MixerSettings[i].KeyName );

                        // query the value key
                        ntStatus = SettingsKey->QueryValueKey( &KeyName,
                                                               KeyValuePartialInformation,
                                                               KeyInfo,
                                                               sizeof(KEY_VALUE_PARTIAL_INFORMATION) + sizeof(DWORD),
                                                               &ResultLength );
                        if(NT_SUCCESS(ntStatus))
                        {
                            PKEY_VALUE_PARTIAL_INFORMATION PartialInfo = PKEY_VALUE_PARTIAL_INFORMATION(KeyInfo);

                            if(PartialInfo->DataLength == sizeof(DWORD))
                            {
                                // set mixer register to registry value
                                MixerRegWrite( MixerSettings[i].RegisterIndex,
                                               BYTE(*(PDWORD(PartialInfo->Data))) );
                            }
                        } else
                        {
                            // if key access failed, set to default
                            MixerRegWrite( MixerSettings[i].RegisterIndex,
                                           MixerSettings[i].DefaultSetting );
                        }
                    }

                    // free the key info
                    ExFreePool(KeyInfo);
                } else
                {
                    // copy default settings
                    for(ULONG i = 0; i < SIZEOF_ARRAY(MixerSettings); i++)
                    {
                        MixerRegWrite( MixerSettings[i].RegisterIndex,
                                       MixerSettings[i].DefaultSetting );
                    }

                    ntStatus = STATUS_INSUFFICIENT_RESOURCES;
                }
            }
            // release the settings key
            SettingsKey->Release();
        }
        // release the driver key
        DriverKey->Release();
    }

    return ntStatus;
}

/*****************************************************************************
 * CAdapterCommon::SaveMixerSettingsToRegistry()
 *****************************************************************************
 * Saves the mixer settings to the registry.
 */
STDMETHODIMP
CAdapterCommon::
SaveMixerSettingsToRegistry
(   void
)
{
    PREGISTRYKEY    DriverKey;
    PREGISTRYKEY    SettingsKey;

    _DbgPrintF(DEBUGLVL_VERBOSE,("[SaveMixerSettingsToRegistry]"));

    // open the driver registry key
    NTSTATUS ntStatus = PcNewRegistryKey( &DriverKey,               // IRegistryKey
                                          NULL,                     // OuterUnknown
                                          DriverRegistryKey,        // Registry key type
                                          KEY_ALL_ACCESS,           // Access flags
                                          m_pDeviceObject,          // Device object
                                          NULL,                     // Subdevice
                                          NULL,                     // ObjectAttributes
                                          0,                        // Create options
                                          NULL );                   // Disposition
    if(NT_SUCCESS(ntStatus))
    {
        UNICODE_STRING  KeyName;

        // make a unicode strong for the subkey name
        RtlInitUnicodeString( &KeyName, L"Settings" );

        // open the settings subkey
        ntStatus = DriverKey->NewSubKey( &SettingsKey,              // Subkey
                                         NULL,                      // OuterUnknown
                                         KEY_ALL_ACCESS,            // Access flags
                                         &KeyName,                  // Subkey name
                                         REG_OPTION_NON_VOLATILE,   // Create options
                                         NULL );
        if(NT_SUCCESS(ntStatus))
        {
            // loop through all mixer settings
            for(UINT i = 0; i < SIZEOF_ARRAY(MixerSettings); i++)
            {
                // init key name
                RtlInitUnicodeString( &KeyName, MixerSettings[i].KeyName );

                // set the key
                DWORD KeyValue = DWORD(MixerSettings[i].CurrentSetting);
                ntStatus = SettingsKey->SetValueKey( &KeyName,                 // Key name
                                                     REG_DWORD,                // Key type
                                                     PVOID(&KeyValue),
                                                     sizeof(DWORD) );
                if(!NT_SUCCESS(ntStatus))
                {
                    break;
                }
            }

            // release the settings key
            SettingsKey->Release();
        }

        // release the driver key
        DriverKey->Release();

    }

    return ntStatus;
}

/*****************************************************************************
 * CAdapterCommon::PowerChangeState()
 *****************************************************************************
 * Change power state for the device.
 */
STDMETHODIMP_(void)
CAdapterCommon::
PowerChangeState
(
    IN      POWER_STATE     NewState
)
{
    UINT i;

    _DbgPrintF( DEBUGLVL_VERBOSE, ("[CAdapterCommon::PowerChangeState]"));

    // is this actually a state change??
    if( NewState.DeviceState != m_PowerState )
    {
        // switch on new state
        switch( NewState.DeviceState )
        {
            case PowerDeviceD0:
                // Insert your code here for entering the full power state (D0).
                // This code may be a function of the current power state.  Note that
                // property accesses such as volume and mute changes may occur when
                // the device is in a sleep state (D1-D3) and should be cached in
                // the driver to be restored upon entering D0.  However, it should
                // also be noted that new miniport and new streams will only be
                // attempted at D0 (portcls will place the device in D0 prior to the
                // NewStream call).

                // Save the new state.  This local value is used to determine when to cache
                // property accesses and when to permit the driver from accessing the hardware.
                m_PowerState = NewState.DeviceState;

#if 0
                // restore mixer settings
                for(i = 0; i < DSP_MIX_MAXREGS - 1; i++)
                {
                    if( i != DSP_MIX_MICVOLIDX )
                    {
                        MixerRegWrite( BYTE(i), MixerSettings[i] );
                    }
                }
#endif
            case PowerDeviceD1:
                // This sleep state is the lowest latency sleep state with respect to the
                // latency time required to return to D0.  The driver can still access
                // the hardware in this state if desired.  If the driver is not being used
                // an inactivity timer in portcls will place the driver in this state after
                // a timeout period controllable via the registry.

            case PowerDeviceD2:
                // This is a medium latency sleep state.  In this state the device driver
                // cannot assume that it can touch the hardware so any accesses need to be
                // cached and the hardware restored upon entering D0 (or D1 conceivably).

            case PowerDeviceD3:
                // This is a full hibernation state and is the longest latency sleep state.
                // The driver cannot access the hardware in this state and must cache any
                // hardware accesses and restore the hardware upon returning to D0 (or D1).

                // Save the new state.
                m_PowerState = NewState.DeviceState;

                _DbgPrintF(DEBUGLVL_VERBOSE,("  Entering D%d",ULONG(m_PowerState)-ULONG(PowerDeviceD0)));
                break;

            default:
                _DbgPrintF(DEBUGLVL_VERBOSE,("  Unknown Device Power State"));
                break;
        }
    }
}

/*****************************************************************************
 * CAdapterCommon::QueryPowerChangeState()
 *****************************************************************************
 * Query to see if the device can
 * change to this power state
 */
STDMETHODIMP_(NTSTATUS)
CAdapterCommon::
QueryPowerChangeState
(
    IN      POWER_STATE     NewStateQuery
)
{
    _DbgPrintF( DEBUGLVL_TERSE, ("[CAdapterCommon::QueryPowerChangeState]"));

    // Check here to see of a legitimate state is being requested
    // based on the device state and fail the call if the device/driver
    // cannot support the change requested.  Otherwise, return STATUS_SUCCESS.
    // Note: A QueryPowerChangeState() call is not guaranteed to always preceed
    // a PowerChangeState() call.

    return STATUS_SUCCESS;
}

/*****************************************************************************
 * CAdapterCommon::QueryDeviceCapabilities()
 *****************************************************************************
 * Called at startup to get the caps for the device.  This structure provides
 * the system with the mappings between system power state and device power
 * state.  This typically will not need modification by the driver.
 *
 */
STDMETHODIMP_(NTSTATUS)
CAdapterCommon::
QueryDeviceCapabilities
(
    IN      PDEVICE_CAPABILITIES    PowerDeviceCaps
)
{
    _DbgPrintF( DEBUGLVL_TERSE, ("[CAdapterCommon::QueryDeviceCapabilities]"));

    return STATUS_SUCCESS;
}

/*****************************************************************************
 * InterruptServiceRoutine()
 *****************************************************************************
 * ISR.
 */
NTSTATUS
InterruptServiceRoutine
(
    IN      PINTERRUPTSYNC  InterruptSync,
    IN      PVOID           DynamicContext
)
{
    NTSTATUS ntStatus = STATUS_UNSUCCESSFUL;

    ASSERT(InterruptSync);
    ASSERT(DynamicContext);

    CAdapterCommon *that = (CAdapterCommon *) DynamicContext;

    //
    // We are here because the MPU tried and failed, so
    // must be a wave interrupt.
    //
    ASSERT(that->m_pAlsBase);
    ASSERT(that->m_pWaveBase);
    ASSERT(that->m_pServiceGroupWave);

    BYTE GcrStatus  = READ_PORT_UCHAR(that->m_pAlsBase + 0x0E);     // read irq status port
    if(GcrStatus & (0x80 | 0x40)){     // playback irq or record irq detected
        if (that->m_pPortWave){
            //
            // Tell it it needs to do some work.
            //
            that->m_pPortWave->Notify(that->m_pServiceGroupWave);
        }
        if(GcrStatus & 0x80){           // playback irq
            READ_PORT_UCHAR(that->m_pWaveBase + DSP_REG_ACK16BIT);  // ack
//            READ_PORT_UCHAR(that->m_pWaveBase + DSP_REG_ACK8BIT);  // ack
        }
        if(GcrStatus & 0x40){           // record irq
            READ_PORT_UCHAR(that->m_pWaveBase + DSP_REG_RESET);     // ack
        }
        ntStatus = STATUS_SUCCESS;
    }
    WRITE_PORT_UCHAR(that->m_pAlsBase + 0x0E, GcrStatus);           // clear irq status flag

    return ntStatus;
}
